"""
    This module is a wrapper for C-types that runs the C-function based on c-type function object and parameter map.
"""

import ctypes
import re
import numpy as np

# Regex pattern to determine if c-type is a primative or array
isArray = re.compile(r"<class '.*\.LP_(c_.*)'>")


def array_to_list(input_array, size):
    """ converts ctypes array to Python list """
    return [input_array[i] for i in range(size)]


def list_to_array(floats, type):
    """ convert Python list to ctypes array """
    if floats is None:
        return None
    data_length_type = type * len(floats)
    return ctypes.cast(data_length_type(*floats), ctypes.POINTER(type))


def update_boundary_mask(input_list, bad):
    """ update boundary mask for new invalid entries that were replaced """
    mask = []
    for i in range(len(input_list)):
        if input_list[i] == np.float32(bad):
            mask.append(True)
        else:
            mask.append(False)
    return mask


def run_solo_function(c_func, args):
    """ 
        Runs a solo library function from python
        
        Args:
            c_func: a reference to the solo function, from c_types
            args: a dictionary of {String : DataTypeValue}
                String is the name of the parameter as it's called from the function
                DataTypeValue contains c-types type and corresponding value.

        Returns:
          Numpy masked array: Contains an array of data, mask, and fill_value of results.

    """

    # data list either comes from "data" or "data1" parameters.
    # data is 1 ray with n gates.
    if "data" not in args:
        if "data1" in args:
            input_list_data = args["data1"].value
        else:
            # All solo functions have data or data1. If not in parametrs, raise error.
            raise ValueError(
                "Expected either Data or Data1 as a parameter for solo functions."
            )
    else:
        input_list_data = args["data"].value

    if "bad" in args:
        bad = args["bad"].value
    else:
        bad = 0
        
    # Obtain clipping and boundary mask, if it's none, then change it to defaults.
    dgi_clip_gate = args["dgi_clip_gate"].value
    boundary_mask = args["boundary_mask"].value

    # retrieve size of input/output/mask array
    data_length = len(input_list_data)

    # nGates will always be the size of data_length, so update its value in args dict.
    args["nGates"].value = data_length

    # optional parameters, if None, then set to default values.
    if boundary_mask is None:
        args["boundary_mask"].value = [True] * data_length # setting boundary mask to all true here will mean C-function does operation on entire dataset. 
    if dgi_clip_gate is None:
        args["dgi_clip_gate"].value = data_length # clip to the end of the data

    # newData is an array expected from solo functions that eventually gets filled with values. 
    # newData must be equal in size as input_data. Initialize it with copying input_data to ensure they are
    # the same size. Not all solo functions have newData parameter.
    if "newData" in args:
        args["newData"].value = input_list_data.copy()

    # parallel lists below
    argtypes = [] # becomes a list of c types expected from C-function call, in order that they are called
    parameters = [] # becomes a list of values, these values correspond to the types in argtypes.

    for i, j in args.items(): # i = key (string of parameter), j = value (DataTypeValue structure with 'type' and 'value')
        argtypes.append(j.type) # extract c-type & append
        result = isArray.search(str(j.type)) # check if this type is an array
        if result: # if so...
            array_type = result.group(1) # extract what type of array... either a float or bool array.
            if (array_type == 'c_float'):
                array = list_to_array(j.value, ctypes.c_float) # convert list to array by specifying the list's values and designated type.
            elif (array_type == 'c_bool'):
                array = list_to_array(j.value, ctypes.c_bool)
            else: # solo functions don't have any other array tile than floats/bools.
                raise Exception("Unexpected type for array.")
            args[i].value = array # finally, update this list to array conversion onto dictionary
            parameters.append(j.value) # and add this array to the parameter list
        else:
            parameters.append(j.type(j.value)) # if not array, simply add to parameter list

    # none of the solo functions have return types
    c_func.restype = None
    # obtained from iteration above
    c_func.argtypes = argtypes
    # run the actual function here. 
    c_func(*parameters)

    # running c_func either:
    # updated "newData" with... new data
    # or
    # updated "bad_flag_mask" with new masks
    if "newData" in args: # if newData was updated...
        # convert resultant ctypes array back to python list
        output_list = array_to_list(args['newData'].value, data_length)
        # create and return masked array
        output_mask = update_boundary_mask(output_list, bad)
        return np.ma.masked_array(data=output_list, mask=output_mask, fill_value=bad)
    elif "bad_flag_mask" in args:
        output_flag_list = array_to_list(args['bad_flag_mask'].value, data_length)
        return np.ma.masked_array(data=input_list_data, mask=output_flag_list, fill_value=bad)
    raise Exception("Unexpected control flow.")
    


# this module handles mapping of C-function names
# also handles which library to import based on platform

import platform
from pathlib import Path
import ctypes

aliases = {}

# list of functions PySolo uses
functions = [
    "assert_bad_flags",
    "assign_value",
    "bad_flags_logic", 
    "BB_unfold_first_good_gate",
    "BB_unfold_local_wind",
    "clear_bad_flags",
    "copy_bad_flags", 
    "do_clear_bad_flags_array",
    "despeckle",
    "flagged_add", 
    "flag_freckles",
    "flag_glitches",
    "funfold",
    "get_boundary_mask",
    "merge_fields",
    "radial_shear",
    "rain_rate",
    "ring_zap",
    "set_bad_flags",
    "threshold_field",
    "remove_ac_motion",
    "remove_storm_motion",
]

# from this script file, go up two directories (pysolo) then into libs/libSolo...
pysolo_dir = Path(__file__).parents[1].absolute()

# get appropriate library depending on platform
# the DLL from Windows has "extern C" so no name mangling occurs.
if (platform.system() == "Windows"):
    path_to_file = pysolo_dir / Path('libs/solo.dll')
    c_lib = ctypes.CDLL(str(path_to_file))
    for function in functions:
        aliases[function] = c_lib["se_" + function]

# Linux, on the other hand, has name manging.
# Algorithm was required to retrieve the functions from the mangled names.
else:
    import os
    import re

    temp_dir = pysolo_dir / Path("temp")
    shared_lib_path = Path(__file__).parents[1].absolute() / Path('libs/libSoloNew.so')

    # run readelf to get a list of C-functions with their mangled names, save results to file
    os.system("mkdir %s" % temp_dir)
    os.system("readelf -Ws %s > %s" % (shared_lib_path, temp_dir / Path("readelf.txt")))

    # read all content from the command
    content = open(temp_dir / Path("readelf.txt")).read()

    # use regex pattern to discover all mangled functions, retrieve the unmangled name from group
    matches = re.findall(r'(_Z\w+se_\w+)', content, re.M)

    # initialize c_types object for shared library
    c_lib = ctypes.CDLL(str(shared_lib_path))
    # iterate through matches, check if PySolo uses that function. If so, save to map.
    for match in matches:
        for func in functions:
            if func in match:
                aliases[func] = c_lib[match]

    # remove command output.
    os.system("rm -r %s" % temp_dir)

# make sure all the needed functions are mapped
for func in functions:
    assert func in aliases, func



import ctypes

from ..c_wrapper.run_solo import run_solo_function

from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases

se_rain_rate = aliases['rain_rate']

def rain_rate(input_list_data, bad, d_const, dgi_clip_gate=None, boundary_mask=None):
    """ 
        Calculates rain_rate in m/s from reflectivity.
        
        Args:
            input_list: A list containing float data,
            bad: A float that represents a missing/invalid data point,
            d_const: A constant value used from the rain rate calculation.
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list)
            (optional) boundary_mask: Defines region over which operations will be done. (default: all True).

        Returns:
          Numpy masked array: Contains an array of data, mask, and fill_value of results.

        Throws:
          ValueError: if input_list and input_boundary_mask are not equal in size,
    """

    args = {
        "d_const" : DataPair.DataTypeValue(ctypes.c_float, d_const),
        "data" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data),
        "newData" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), None),
        "nGates" : DataPair.DataTypeValue(ctypes.c_size_t, None),
        "bad" : DataPair.DataTypeValue(ctypes.c_float, bad),
        "dgi_clip_gate" : DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
    }

    return run_solo_function(se_rain_rate, args)


def rain_rate_masked(masked_array, d_const, boundary_mask=None):
    """ 
        Performs a <TODO> operation on a numpy masked array
        
        Args:
            masked_array: A numpy masked array data structure,
            d_const: A constant value used from the rain rate calculation.

        Returns:
            Numpy masked array

        Throws:
            ModuleNotFoundError: if numpy is not installed
            AttributeError: if masked_array arg is not a numpy masked array.
    """


    return masked_op.masked_func(rain_rate, masked_array, d_const, boundary_mask = boundary_mask)



import ctypes

from ..c_wrapper.run_solo import run_solo_function
from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases

se_merge_fields = aliases['merge_fields']

def merge_fields(input_list_data_1, input_list_data_2, bad, dgi_clip_gate=None, boundary_mask=None):
    """ 
        Replaces bad values from input_list_data_1 from values in input_list_data2
        
        Args:
            input_list_data_1: The list of floats to be modified,
            input_list_data_2: The list of floats that replaces values from input_list_data_1,
            bad: A float that represents a missing/invalid data point,
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list)
            (optional) boundary_mask: Defines region over which operations will be done. (default: all True).

        Returns:
          Numpy masked array: Contains an array of data, mask, and fill_value of results.

        Throws:
          ValueError: if input_list and input_boundary_mask are not equal in size,

    """

    args = {
        "data1" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data_1),
        "data2" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data_2),
        "newData" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), None),
        "nGates" : DataPair.DataTypeValue(ctypes.c_size_t, None),
        "bad" : DataPair.DataTypeValue(ctypes.c_float, bad),
        "dgi_clip_gate" : DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
    }

    return run_solo_function(se_merge_fields, args)


def merge_fields_masked(masked_array, reference_masked_array, boundary_mask=None):
    """ 
        Replaces bad values from input_list_data_1 from values in input_list_data2
        
        Args:
            masked_array: A numpy masked array that is to be modified.
            reference_masked_array: A numpy masked array with values that replace masked_array on bad entries.

        Returns:
            Numpy masked array

        Throws:
            ModuleNotFoundError: if numpy is not installed
            AttributeError: if masked_array arg is not a numpy masked array.
    """
    return masked_op.masked_func(merge_fields, masked_array,  boundary_mask = boundary_mask, second_masked_array=reference_masked_array)



'''
    This module is a wrapper for run_solo. Allows use of run_solo with 2D masked arrays rather than 1D arrays.
'''

import numpy as np


def masked_func(func, masked_array, *args, boundary_mask=None, second_masked_array=None, usesBadFlags=False):
    """ 
        Performs a solo function on a 2D masked array.
        
        Args:
            masked_array: A numpy masked array data structure with input data (usually this contains info on a single field),
            args: A list of args, listed in order, as required by the solo function
            (optional) boundary_mask: A boolean list designating which region to perform the operation. (Default = all True, so entire region is operated on.)
            (optional) second_masked_array: Some solo functions do operations on two fields. If so, assign this parameter to the masked array of that other field.
            (optional) usesBadFlags: Some solo functions do operations on masks rather than data. If so, set to true.

        Returns:
            Numpy masked array

        Throws:
            AttributeError: if masked_array arg is not a numpy masked array.
    """


    if isinstance(masked_array, np.ma.masked_array):
        raise AttributeError("Expected a numpy masked array type for 'masked_array' parameter, received ", type(masked_array))

    missing = masked_array.fill_value
    mask = masked_array.mask.tolist()
    if not mask:
        mask = [False] * len(data_list)
    data_list = masked_array.tolist(missing)

    
    # initialize lists with data/masks. These will become lists of lists
    output_data = []
    output_mask = []

    # iterate through each ray
    for i in range(len(data_list)):
        input_data = data_list[i] # gates
        input_mask = mask[i] # mask for gates

        if second_masked_array is not None: # if using second masked array, obtain gates for that
            second_missing = second_masked_array.fill_value
            second_input_data = second_masked_array.tolist(second_missing)[i]
            # in pysolo, second parameter of functions always designates data list for secondary masked list, for functions that have secondary masked arrays.
            func_ma = func(input_data, second_input_data, missing, *args, boundary_mask=boundary_mask)
        elif usesBadFlags:
            func_ma = func(input_data, missing, *args, input_mask, boundary_mask=boundary_mask)
        else:
            func_ma = func(input_data, missing, *args, boundary_mask=boundary_mask)

        output_data.append(np.ma.getdata(func_ma)) # add data to output list of lists
        output_mask.append(np.ma.getdata(func_ma.mask)) # add mask as well

    # convert the list of lists, to a 2D masked array
    output_masked_array = np.ma.masked_array(data=output_data, mask=output_mask, fill_value=missing)
    return output_masked_array


import ctypes
from ..c_wrapper.run_solo import run_solo_function

from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases

se_unfold_local_wind = aliases['BB_unfold_local_wind']

def unfold_local_wind(input_list_data, bad, nyquist_velocity, dds_radd_eff_unamb_vel, azimuth_angle_degrees, elevation_angle_degrees, ew_wind,  ns_wind,  ud_wind, max_pos_folds, max_neg_folds, ngates_averaged, dgi_clip_gate=None, boundary_mask=None):
    """
        Performs a <TODO>

        Args:
            input_list: A list containing float data,
            bad: A float that represents a missing/invalid data point,
            nyquist_velocity: <TODO>,
            dds_radd_eff_unamb_vel: <TODO>,
            azimuth_angle_degrees: <TODO>,
            elevation_angle_degrees: <TODO>,
            ew_wind: <TODO>,
            ns_wind: <TODO>,
            ud_wind: <TODO>,
            max_pos_folds: <TODO>,
            max_neg_folds: <TODO>,
            ngates_averaged: <TODO>,
            last_good_v0: <TODO>,
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list)
            (optional) boundary_mask: Defines region over which operations will be done. (default: all True).

        Returns:
          Numpy masked array: Contains an array of data, mask, and fill_value of results.

        Throws:
          ValueError: if input_list and input_boundary_mask are not equal in size,
    """

    args = {
        "data" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data),
        "newData" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), None),
        "nGates" : DataPair.DataTypeValue(ctypes.c_size_t, None),
        "nyquist_velocity" : DataPair.DataTypeValue(ctypes.c_float, nyquist_velocity),
        "dds_radd_eff_unamb_vel" : DataPair.DataTypeValue(ctypes.c_float, dds_radd_eff_unamb_vel),
        "azimuth_angle_degrees" : DataPair.DataTypeValue(ctypes.c_float, azimuth_angle_degrees),
        "elevation_angle_degrees" : DataPair.DataTypeValue(ctypes.c_float, elevation_angle_degrees),
        "ew_wind" : DataPair.DataTypeValue(ctypes.c_float, ew_wind),
        "ns_wind" : DataPair.DataTypeValue(ctypes.c_float, ns_wind),
        "ud_wind" : DataPair.DataTypeValue(ctypes.c_float, ud_wind),
        "max_pos_folds" : DataPair.DataTypeValue(ctypes.c_int, max_pos_folds),
        "max_neg_folds" : DataPair.DataTypeValue(ctypes.c_int, max_neg_folds),
        "ngates_averaged" : DataPair.DataTypeValue(ctypes.c_size_t, ngates_averaged),
        "bad" : DataPair.DataTypeValue(ctypes.c_float, bad),
        "dgi_clip_gate" : DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
    }

    return run_solo_function(se_unfold_local_wind, args)


def unfold_local_wind_masked(masked_array, nyquist_velocity, dds_radd_eff_unamb_vel, azimuth_angle_degrees, elevation_angle_degrees, ew_wind, ns_wind, ud_wind, max_pos_folds, max_neg_folds, ngates_averaged, boundary_mask=None):
    """
        Performs a <TODO> on a numpy masked array

        Args:
            masked_array: A numpy masked array data structure,
            nyquist_velocity: <TODO>
            dds_radd_eff_unamb_vel: <TODO>
            max_pos_folds: <TODO>
            max_neg_folds: <TODO>
            ngates_averaged: <TODO>
            last_good_v0: <TODO>

        Returns:
            Numpy masked array

        Throws:
            ModuleNotFoundError: if numpy is not installed
            AttributeError: if masked_array arg is not a numpy masked array.
    """
    
    return masked_op.masked_func(unfold_local_wind, masked_array, nyquist_velocity, dds_radd_eff_unamb_vel, azimuth_angle_degrees, elevation_angle_degrees, ew_wind, ns_wind, ud_wind, max_pos_folds, max_neg_folds, ngates_averaged, boundary_mask = boundary_mask)



from ...c_wrapper.run_solo import run_solo_function, list_to_array, array_to_list
from ...c_wrapper import DataPair, masked_op
from ...c_wrapper.function_alias import aliases
from ...enums import Where



import ctypes
from ..c_wrapper.run_solo import run_solo_function

from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases

se_flag_glitches = aliases['flag_glitches']

def flag_glitches(input_list_data, bad, deglitch_threshold, deglitch_radius, deglitch_min_gates, bad_flag_mask, dgi_clip_gate=None, boundary_mask=None):
    """
        Routine to remove discountinuities (freckles) from the data.

        Args:
            input_list: A list containing float data.
            bad: A float that represents a missing/invalid data point.
            deglitch_threshold: <TODO>
            deglitch_radius: <TODO>
            deglitch_min_gates: <TODO>
            bad_flag_mask: A mask for input_list marking good or bad values.
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list)
            (optional) boundary_mask: Defines region over which operations will be done. (default: all True)

        Returns:
            Numpy masked array: Contains an array of data, mask, and fill_value of results.


    """

    args = {
        "deglitch_threshold" : DataPair.DataTypeValue(ctypes.c_float, deglitch_threshold),
        "deglitch_radius" : DataPair.DataTypeValue(ctypes.c_int, deglitch_radius),
        "deglitch_min_gates" : DataPair.DataTypeValue(ctypes.c_int, deglitch_min_gates),
        "data" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data),
        "nGates" : DataPair.DataTypeValue(ctypes.c_size_t, None),
        "bad" : DataPair.DataTypeValue(ctypes.c_float, bad),
        "dgi_clip_gate" : DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
        "bad_flag_mask" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), bad_flag_mask)
    }

    return run_solo_function(se_flag_glitches, args)


def flag_glitches_masked(masked_array, deglitch_threshold, deglitch_radius, deglitch_min_gates, boundary_mask=None):
    """ 
        Routine to remove discountinuities (freckles) from the data.
        
        Args:
            masked_array: A numpy masked array data structure,
            bad_flag_mask: A list of lists,
            deglitch_threshold: <TODO>,
            deglitch_radius: <TODO>,
            deglitch_min_gates: <TODO>

        Returns:
            Numpy masked array

        Throws:
            ModuleNotFoundError: if numpy is not installed
            AttributeError: if masked_array arg is not a numpy masked array.
    """

    return masked_op.masked_func(flag_glitches, masked_array, deglitch_threshold, deglitch_radius, deglitch_min_gates, boundary_mask = boundary_mask, usesBadFlags=True)



import ctypes

from ..c_wrapper.run_solo import run_solo_function
from ..c_wrapper import DataPair, masked_op
from ..c_wrapper.function_alias import aliases

se_ring_zap = aliases['ring_zap']

def ring_zap(input_list_data, bad, from_km, to_km, km_between_gates=1, dgi_clip_gate=None, boundary_mask=None):
    """ 
        Performs a ring zap operation on a list of data.
        
        Args:
            input_list: A list containing float data,
            bad: A float that represents a missing/invalid data point,
            from_km: An integer for the starting range,
            to_km: An integer for the ending range,
            (optional) km_between_gates: An integer representing the distance (in km) between gates (default: 1 km).
            (optional) dgi_clip_gate: An integer determines the end of the ray (default: length of input_list).
            (optional) boundary_mask: Defines region over which operations will be done (default: all True).

        Returns:
          Numpy masked array: Contains an array of data, mask, and fill_value of results.

    """

    from_km = int(from_km / km_between_gates)
    to_km = int(to_km / km_between_gates)

    args = {
        "from_km" : DataPair.DataTypeValue(ctypes.c_size_t, from_km),
        "to_km" : DataPair.DataTypeValue(ctypes.c_size_t, to_km),
        "data" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), input_list_data),
        "newData" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_float), None),
        "nGates" : DataPair.DataTypeValue(ctypes.c_size_t, None),
        "bad" : DataPair.DataTypeValue(ctypes.c_float, bad),
        "dgi_clip_gate" : DataPair.DataTypeValue(ctypes.c_size_t, dgi_clip_gate),
        "boundary_mask" : DataPair.DataTypeValue(ctypes.POINTER(ctypes.c_bool), boundary_mask),
    }

    return run_solo_function(se_ring_zap, args)


def ring_zap_masked(masked_array, from_km, to_km, km_between_gates, boundary_mask=None):
    """ 
        Performs a ring zap operation on a numpy masked array
        
        Args:
            masked_array: A numpy masked array data structure,
            from_km: An integer for the starting range,
            to_km: An integer for the ending range,
            km_between_gates: An integer representing the distance (in km) between gates

        Returns:
            Numpy masked array

        Throws:
            ModuleNotFoundError: if numpy is not installed
            AttributeError: if masked_array arg is not a numpy masked array.
    """

    from_km = int(from_km / km_between_gates)
    to_km = int(to_km / km_between_gates)

    return masked_op.masked_func(ring_zap, masked_array, from_km, to_km, boundary_mask = boundary_mask)
